use super::{
    modifiers::{GroupType, MatchType, Mod},
    Operable,
};
use crate::{query::IntoQuery, seal::Sealed};
use std::fmt::Display;

/// Represents a Comparison operation between type `A` and type `B`
/// where `A` and `B` both implement [``Operable``]
///
/// This struct is not meant to be constructed directly by the user of the library.
/// Queries with arithmetic operations should be constructed through the [``Comparison``] trait.
#[derive(Debug, Clone)]
pub struct CompOperation<'a, A, B>
where
    A: Operable,
    B: Operable,
{
    cmp_type: CmpOp,
    bool: bool,
    left: A,
    right: B,
    matching: Mod<'a, MatchType>,
    group_mod: Mod<'a, GroupType>,
}

impl<'a, A, B> CompOperation<'a, A, B>
where
    A: Operable,
    B: Operable,
{
    fn new(cmp_type: CmpOp, bool: bool, left: A, right: B) -> Self {
        Self {
            cmp_type,
            left,
            right,
            bool,
            matching: Default::default(),
            group_mod: Default::default(),
        }
    }

    /// Apply `group_left` group modifier to the comparison operation without specifying any label names
    pub fn group_left(mut self) -> Self {
        self.group_mod = Mod::from((GroupType::Left, []));
        self
    }

    /// Apply `group_left` group modifier to the comparison operation with label names
    pub fn group_left_labels(mut self, labels: impl IntoIterator<Item = &'a str>) -> Self {
        self.group_mod = Mod::from((GroupType::Left, labels));
        self
    }

    /// Apply `group_right` group modifier to the comparison operation without specifying any label names
    pub fn group_right(mut self) -> Self {
        self.group_mod = Mod::from((GroupType::Right, []));
        self
    }

    /// Apply `group_right` group modifier to the comparison operation with label names
    pub fn group_right_labels(mut self, labels: impl IntoIterator<Item = &'a str>) -> Self {
        self.group_mod = Mod::from((GroupType::Right, labels));
        self
    }

    /// Match only given labels between the two vectors for evauluating the comparison operation
    ///
    /// Note: Only to be used with vector expressions on both left and right
    /// i.e. `<vector expr> <bin-op> on(<label list>) <vector expr>`
    pub fn match_on(mut self, labels: impl IntoIterator<Item = &'a str>) -> Self {
        self.matching = Mod::from((MatchType::On, labels));
        self
    }

    /// Match only given labels between the two vectors for evauluating the comparison operation
    ///
    /// Note: Only to be used with vector expressions on both left and right
    /// i.e. `<vector expr> <bin-op> ignoring(<label list>) <vector expr>`
    pub fn ignoring(mut self, labels: impl IntoIterator<Item = &'a str>) -> Self {
        self.matching = Mod::from((MatchType::Ignoring, labels));
        self
    }
}

impl<A, B> Display for CompOperation<'_, A, B>
where
    A: Operable,
    B: Operable,
{
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let b = self.bool.then_some(" bool").unwrap_or_default();
        write!(
            f,
            "({left} {op}{b}{matching}{group} {right})",
            left = self.left,
            right = self.right,
            op = self.cmp_type.to_str(),
            matching = self.matching,
            group = self.group_mod,
        )
    }
}

impl<A, B> Operable for CompOperation<'_, A, B>
where
    A: Operable,
    B: Operable,
{
}

impl<A, B> Sealed for CompOperation<'_, A, B>
where
    A: Operable,
    B: Operable,
{
}

impl<A, B> IntoQuery for CompOperation<'_, A, B>
where
    A: Operable,
    B: Operable,
{
    type Target = String;
    fn into_query(self) -> Self::Target {
        self.to_string()
    }
}

#[derive(Debug, Clone, Copy)]
enum CmpOp {
    Eq,
    Ne,
    Gt,
    Lt,
    Gteq,
    Lteq,
}

impl CmpOp {
    fn to_str(self) -> &'static str {
        match self {
            CmpOp::Eq => "==",
            CmpOp::Ne => "!=",
            CmpOp::Gt => ">",
            CmpOp::Lt => "<",
            CmpOp::Gteq => ">=",
            CmpOp::Lteq => "<=",
        }
    }
}

/// Trait to enable comparison query operators on all types that implement [``Operable``]
pub trait Comparison<'a>: Operable
where
    Self: Sized,
{
    /// Apply the equality or `==` comparison operator to `self` and another operable type
    ///
    /// The resulting query will have a format of `(self == right)`
    fn cmp_eq<R: Operable>(self, right: R) -> CompOperation<'a, Self, R> {
        CompOperation::new(CmpOp::Eq, false, self, right)
    }

    /// Apply the equality or `==` comparison operator to `self` and another operable type with the bool modifier
    ///
    /// The resulting query will have a format of `(self == bool right)`
    fn cmp_eq_bool<R: Operable>(self, right: R) -> CompOperation<'a, Self, R> {
        CompOperation::new(CmpOp::Eq, true, self, right)
    }

    /// Apply the inequality or `!=` comparison operator to `self` and another operable type
    ///
    /// The resulting query will have a format of `(self == right)`
    fn cmp_ne<R: Operable>(self, right: R) -> CompOperation<'a, Self, R> {
        CompOperation::new(CmpOp::Ne, false, self, right)
    }

    /// Apply the inequality or `!=` comparison operator to `self` and another operable type with the bool modifier
    ///
    /// The resulting query will have a format of `(self != bool right)`
    fn cmp_ne_bool<R: Operable>(self, right: R) -> CompOperation<'a, Self, R> {
        CompOperation::new(CmpOp::Ne, true, self, right)
    }

    /// Apply the greater-than or `>` comparison operator to `self` and another operable type
    ///
    /// The resulting query will have a format of `(self > right)`
    fn cmp_gt<R: Operable>(self, right: R) -> CompOperation<'a, Self, R> {
        CompOperation::new(CmpOp::Gt, false, self, right)
    }

    /// Apply the greater-than or `>` comparison operator to `self` and another operable type with the bool modifier
    ///
    /// The resulting query will have a format of `(self > bool right)`
    fn cmp_gt_bool<R: Operable>(self, right: R) -> CompOperation<'a, Self, R> {
        CompOperation::new(CmpOp::Gt, true, self, right)
    }

    /// Apply the less-than or `<` comparison operator to `self` and another operable type
    ///
    /// The resulting query will have a format of `(self < right)`
    fn cmp_lt<R: Operable>(self, right: R) -> CompOperation<'a, Self, R> {
        CompOperation::new(CmpOp::Lt, false, self, right)
    }

    /// Apply the less-than or `<` comparison operator to `self` and another operable type with the bool modifier
    ///
    /// The resulting query will have a format of `(self < bool right)`
    fn cmp_lt_bool<R: Operable>(self, right: R) -> CompOperation<'a, Self, R> {
        CompOperation::new(CmpOp::Lt, true, self, right)
    }

    /// Apply the greater-than/equal or `>=` comparison operator to `self` and another operable type with the bool modifier
    ///
    /// The resulting query will have a format of `(self >= right)`
    fn cmp_gteq<R: Operable>(self, right: R) -> CompOperation<'a, Self, R> {
        CompOperation::new(CmpOp::Gteq, false, self, right)
    }

    /// Apply the greater-than/equal or `>=` comparison operator to `self` and another operable type with the bool modifier
    ///
    /// The resulting query will have a format of `(self >= bool right)`
    fn cmp_gteq_bool<R: Operable>(self, right: R) -> CompOperation<'a, Self, R> {
        CompOperation::new(CmpOp::Gteq, true, self, right)
    }

    /// Apply the less-than/equal or `<=` comparison operator to `self` and another operable type
    ///
    /// The resulting query will have a format of `(self <= bool right)`
    fn cmp_lteq<R: Operable>(self, right: R) -> CompOperation<'a, Self, R> {
        CompOperation::new(CmpOp::Lteq, false, self, right)
    }

    /// Apply the less-than/equal or `<=` comparison operator to `self` and another operable type with the bool modifier
    ///
    /// The resulting query will have a format of `(self <= bool right)`
    fn cmp_lteq_bool<R: Operable>(self, right: R) -> CompOperation<'a, Self, R> {
        CompOperation::new(CmpOp::Lteq, true, self, right)
    }
}

impl<T> Comparison<'_> for T where T: Operable {}

#[cfg(test)]
mod tests {
    use super::*;
    use crate::query::{Metric, Scalar};

    #[test]
    fn cmp_all_ops() {
        let query = Scalar(1.0);
        assert_eq!(query.cmp_eq(query).to_string(), "(1 == 1)");
        assert_eq!(query.cmp_eq_bool(query).to_string(), "(1 == bool 1)");
        assert_eq!(query.cmp_ne(query).to_string(), "(1 != 1)");
        assert_eq!(query.cmp_ne_bool(query).to_string(), "(1 != bool 1)");
        assert_eq!(query.cmp_gt(query).to_string(), "(1 > 1)");
        assert_eq!(query.cmp_gt_bool(query).to_string(), "(1 > bool 1)");
        assert_eq!(query.cmp_lt(query).to_string(), "(1 < 1)");
        assert_eq!(query.cmp_lt_bool(query).to_string(), "(1 < bool 1)");
        assert_eq!(query.cmp_gteq(query).to_string(), "(1 >= 1)");
        assert_eq!(query.cmp_gteq_bool(query).to_string(), "(1 >= bool 1)");
        assert_eq!(query.cmp_lteq(query).to_string(), "(1 <= 1)");
        assert_eq!(query.cmp_lteq_bool(query).to_string(), "(1 <= bool 1)");
    }

    #[test]
    fn cmp_scalar_scalar() {
        let query = Scalar(0.0).cmp_eq(Scalar(1.0));
        assert_eq!(query.to_string(), "(0 == 1)");
    }

    #[test]
    fn cmp_metric_scalar() {
        let query = Metric::new("test_metric").cmp_ne(Scalar(0.5));
        assert_eq!(query.to_string(), "(test_metric != 0.5)");
    }

    #[test]
    fn cmp_nested() {
        let query = Metric::new("test_metric")
            .cmp_eq(Scalar(0.5))
            .cmp_gteq(Scalar(0.1));

        assert_eq!(query.to_string(), "((test_metric == 0.5) >= 0.1)");
    }

    #[test]
    fn cmp_group_left() {
        let metric = Metric::new("test_metric")
            .cmp_eq(Metric::new("another_metric"))
            .group_left()
            .to_string();
        assert_eq!(metric, "(test_metric == group_left another_metric)");
    }

    #[test]
    fn cmp_group_right() {
        let metric = Metric::new("test_metric")
            .cmp_eq(Metric::new("another_metric"))
            .group_right()
            .to_string();
        assert_eq!(metric, "(test_metric == group_right another_metric)");
    }

    #[test]
    fn cmp_group_left_labels() {
        let metric = Metric::new("test_metric")
            .cmp_eq(Metric::new("another_metric"))
            .group_left_labels(["label", "label2"])
            .to_string();
        assert_eq!(
            metric,
            "(test_metric == group_left(label,label2) another_metric)"
        );
    }

    #[test]
    fn cmp_group_right_labels() {
        let metric = Metric::new("test_metric")
            .cmp_eq(Metric::new("another_metric"))
            .group_right_labels(["label", "label2"])
            .to_string();
        assert_eq!(
            metric,
            "(test_metric == group_right(label,label2) another_metric)"
        );
    }
}
